# I18N-related pylint module
#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Chris Lumens <clumens@redhat.com>
#

import astroid

from pylint.checkers import BaseChecker
from pylint.checkers.strings import StringFormatChecker
from pylint.checkers.logging import LoggingChecker
from pylint.checkers.utils import only_required_for_messages

from copy import copy

translationMethods = frozenset(["_", "N_", "P_", "C_", "CN_", "CP_"])


# Returns a list of the message strings for a given translation method call
def _get_message_strings(node):
    msgstrs = []

    if node.func.name in ("_", "N_") and len(node.args) >= 1:
        if isinstance(node.args[0], astroid.Const):
            msgstrs.append(node.args[0].value)
    elif node.func.name in ("C_", "CN_") and len(node.args) >= 2:
        if isinstance(node.args[1], astroid.Const):
            msgstrs.append(node.args[1].value)
    elif node.func.name == "P_" and len(node.args) >= 2:
        if isinstance(node.args[0], astroid.Const):
            msgstrs.append(node.args[0].value)
        if isinstance(node.args[1], astroid.Const):
            msgstrs.append(node.args[1].value)
    elif node.func.name == "CP_" and len(node.args) >= 3:
        if isinstance(node.args[1], astroid.Const):
            msgstrs.append(node.args[1].value)
        if isinstance(node.args[2], astroid.Const):
            msgstrs.append(node.args[2].value)

    return msgstrs


class IntlChecker(BaseChecker):
    name = "internationalization"
    msgs = {"W9901": ("Found % in a call to a _() method",
                      "found-percent-in-_",
                      "% in a call to one of the _() methods results in incorrect translations"),
            "W9902": ("Found _ call at module/class level",
                      "found-_-in-module-class",
                      "Calling _ at the module or class level results in translations to the wrong language")
           }

    @only_required_for_messages("found-percent-in-_")
    def visit_binop(self, node):
        if node.op != "%":
            return

        curr = node
        while curr.parent:
            if isinstance(curr.parent, astroid.Call) and getattr(curr.parent.func, "name", "") in translationMethods:
                self.add_message("W9901", node=node)
                break

            curr = curr.parent

    @only_required_for_messages("found-_-in-module-class")
    def visit_call(self, node):
        # The first test skips internal functions like getattr.
        if isinstance(node.func, astroid.Name) and node.func.name == "_":
            if isinstance(node.scope(), astroid.Module) or isinstance(node.scope(), astroid.ClassDef):
                self.add_message("W9902", node=node)


# Extend LoggingChecker to check translated logging strings
class IntlLoggingChecker(LoggingChecker):
    name = 'intl-logging'
    msgs = {'W9903': ("Fake message for translated E/W120* checks",
                      "translated-log",
                      "This message is not emitted itself, but can be used to control the display of \
                       logging format messages extended for translated strings")
           }

    @only_required_for_messages('translated-log')
    def visit_call(self, node):
        if len(node.args) >= 1 and isinstance(node.args[0], astroid.Call) and \
                getattr(node.args[0].func, "name", "") in translationMethods:
            for formatstr in _get_message_strings(node.args[0]):
                # Both the node and the args need to be copied so we don't replace args
                # on the original node.
                copynode = copy(node)
                copyargs = copy(node.args)
                copyargs[0] = astroid.Const(formatstr)
                copynode.args = copyargs
                LoggingChecker.visit_call(self, copynode)

    def __init__(self, *args, **kwargs):
        LoggingChecker.__init__(self, *args, **kwargs)

        # FIXME: Remove this hack by something what pylint supports.
        # This hack will avoid crash thanks to the colliding option names. The best
        # solution would be when pylint supports inheritance of the existing checkers.
        self.options = ()


# Extend StringFormatChecker to check translated format strings
class IntlStringFormatChecker(StringFormatChecker):
    name = 'intl-string'
    msgs = {'W9904': ("Fake message for translated E/W130* checks",
                      "translated-format",
                      "This message is not emitted itself, but can be used to control the display of \
                       string format messages extended for translated strings")
           }

    @only_required_for_messages('translated-format')
    def visit_binop(self, node):
        if node.op != '%':
            return

        if isinstance(node.left, astroid.Call) and getattr(node.left.func, "name", "") in translationMethods:
            for formatstr in _get_message_strings(node.left):
                # Create a copy of the node with just the message string as the format
                copynode = copy(node)
                copynode.left = astroid.Const(formatstr)
                StringFormatChecker.visit_binop(self, copynode)


def register(linter):
    """required method to auto register this checker """
    linter.register_checker(IntlChecker(linter))
    linter.register_checker(IntlLoggingChecker(linter))
    linter.register_checker(IntlStringFormatChecker(linter))



#
# translatepo.py: translate strings from data in .po files
#
# Copyright (C) 2013-2014  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: David Shea <dshea@redhat.com>

# This module is helpful for testing translated data. The input is the .po files
# in the po directory, so no gettext compilation or LC_MESSAGES directories are
# necessary.

import os
import re

try:
    import polib
except ImportError:
    print("You need to install the python-polib package to read translations")
    raise


class PODict(object):
    def __init__(self, filename):
        """Create a new dictionary of translations from a po file."""

        self._dict = {}
        self._dict[None] = {}

        pofile = polib.pofile(filename)
        self.metadata = pofile.metadata

        # Rearrange the entries in the pofile to make it easier to look up contexts.
        # _dict will be layed out _dict[msgctxt][msgid] = (translated1, translated2, ...)
        # There may be multiple translations because of plurals
        for entry in pofile.translated_entries():
            if (entry.msgctxt is not None) and (entry.msgctxt not in self._dict):
                self._dict[entry.msgctxt] = {}

            # If this is a plural entry, add entries for both the singular and
            # plural forms so that either can be used for a lookup
            if entry.msgstr_plural:
                xlist = [entry.msgstr_plural[key] for key in entry.msgstr_plural.keys()]
                self._dict[entry.msgctxt][entry.msgid] = xlist
                self._dict[entry.msgctxt][entry.msgid_plural] = xlist
            else:
                self._dict[entry.msgctxt][entry.msgid] = (entry.msgstr,)

    def get(self, key, context=None):
        return self._dict[context][key]


# Return a dictionary of PODict objects for each language in a po directory
def translate_all(podir):
    podicts = {}

    with open(os.path.join(podir, 'LINGUAS'), encoding='utf-8') as linguas:
        for line in linguas.readlines():
            if re.match(r'^#', line):
                continue

            for lang in line.strip().split(" "):
                podicts[lang] = PODict(os.path.join(podir, lang + ".po"))

    return podicts



# yum preconf pylint module
#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Chris Lumens <clumens@redhat.com>
#

import astroid

from pylint.checkers import BaseChecker
from pylint.checkers.utils import only_required_for_messages


class PreconfChecker(BaseChecker):
    name = "Yum preconf"
    msgs = {"W9910": ("Accessing yum.preconf outside of _resetYum",
                      "bad-preconf-access",
                      "Accessing yum.preconf outside of _resetYum will cause tracebacks"),
           }

    @only_required_for_messages("bad-preconf-access")
    def visit_getattr(self, node):
        if node.attrname == "preconf":
            if not isinstance(node.scope(), astroid.FunctionDef) or not node.scope().name == "_resetYum":
                self.add_message("W9910", node=node)


def register(linter):
    """required method to auto register this checker """
    linter.register_checker(PreconfChecker(linter))



#
# pangocheck.py: data and methods for checking pango markup strings
#
# Copyright (C) 2014  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: David Shea <dshea@redhat.com>

import re
from collections import Counter

__all__ = ["markup_nodes", "is_markup", "markup_match"]

# "a" isn't actually pango markup, but GtkLabel uses it
markup_nodes = ["markup", "a", "b", "big", "i", "s", "span", "sub", "sup", "small", "tt", "u"]


# Check to see if a string looks like Pango markup, no validation
def is_markup(test_string):
    if isinstance(test_string, bytes):
        test_string = test_string.decode("utf-8", "ignore")

    return any(re.search(r'<\s*%s(\s|>)' % node_type, test_string)
            for node_type in markup_nodes)


# Verify that the translation of a markup string looks more or less like the original
def markup_match(orig_markup, xlated_markup):
    # Look for tags. Create a count of each kind of tag and a list of attributes.
    # "Don't parse XML with regular expressions" I can hear you saying, but we're
    # not trying to match elements, just pull tag-like substrings out of the string.
    # Figuring out if tags are closed or in the right order is someone else's job.
    def _parse_markup(markup_string):
        name_count = Counter()
        attr_count = Counter()

        for tag in re.findall(r'<[^>]*>', markup_string):
            # Treat everything up to the first space, / or > as the element name
            (name, rest) = re.match(r'<([^\s/>]*)(.*)>', tag).groups()
            name_count[name] += 1

            # Strip the / from the rest of the tag, if present
            if rest.endswith('/'):
                rest = rest[:-1]

            # Make a list of attributes that need to be contained in the other string
            attr_count.update(rest.split())

        return (name_count, attr_count)

    (name_count1, attr_count1) = _parse_markup(orig_markup)
    (name_count2, attr_count2) = _parse_markup(xlated_markup)

    name_list1 = sorted(name_count1.elements())
    name_list2 = sorted(name_count2.elements())
    attr_list1 = sorted(attr_count1.elements())
    attr_list2 = sorted(attr_count2.elements())

    return (name_list1 == name_list2) and (attr_list1 == attr_list2)


# Check that the markup is needed at all.
# The input is a parsed ElementTree of the string '<markup>pango markup goes here</markup>'
# The markup is unnecessary if the only markup in the string surrounds the entire rest of
# the string, meaning that the pango attributes apply to the entire string, and thus
# could be expressed using attribute lists. For example, strings like:
#   <b>Bold text</b>
# or
#   <span foreground="grey"><i>colorful</i></span>
# but not strings like:
#   <span size="small">This string contains <b>internal</b> markup</span>
# that contain markup that must be passed to the translators.
#
# This function returns True if the markup is necessary and False if the markup
# can be discarded and expressed as attribute lists.
def markup_necessary(markup_tree):
    # If the element has no children at all, there is no markup inside and the
    # markup is unnecessary.
    if not len(markup_tree):
        return False

    # If there is more than one child, the markup is necessary
    if len(markup_tree) > 1:
        return True

    # QUICK NOTE FOR PEOPLE EXPECTING ElementTree TO ACT KINDA LIKE DOM 'CUZ LOL
    # ElementTree is kind of weird with respect to handling multiple text children
    # of an Element node. element.text is the text leading up to the first element
    # child, and element[child_idx].tail is the text following the child node that
    # is actually a child of element but isn't a property of element because Python
    # is crazy.
    #
    # A string like "<markup>word1<i>word2</i>word3<empty/>word4</markup>" will result in
    #   tree == <Element 'markup' ...>
    #   tree.text == 'word1'
    #   tree[0] == <Element 'i' ...>
    #   tree[0].text == 'word2'
    #   tree[0].tail == 'word3'
    #   tree[1] == <Element 'empty' ...>
    #   tree[1].text == None
    #   tree[1].text == 'word4'
    #
    # So elements that contain text before a child markup element will have
    # element.text is not None. Elements that have text after a child element
    # will have .tail on that child set to not None.

    # If .text is set, there is text before the child node, as in
    # <span>text <b>child</b></span>
    # and the markup is necessary
    if markup_tree.text:
        return True

    # If the child (we already know there's only one) has .tail set, then
    # there is text between the close of the child and the end of the element
    # and the markup is necessary
    if markup_tree[0].tail:
        return True

    # Recurse on the child node
    return markup_necessary(markup_tree[0])



# setenv pylint module
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Shea <dshea@redhat.com>
#

import astroid

from pylint.checkers import BaseChecker
from pylint.checkers.utils import only_required_for_messages, safe_infer

import os


class EnvironChecker(BaseChecker):
    name = "environ"
    msgs = {"W9940" : ("Found potentially unsafe modification of environment",
                       "environment-modify",
                       "Potentially thread-unsafe modification of environment")}

    def _is_environ(self, node):
        # Guess whether a node being modified is os.environ

        if isinstance(node, astroid.Attribute):
            if node.attrname == "environ":
                expr_node = safe_infer(node.expr)
                if isinstance(expr_node, astroid.Module) and expr_node.name == "os":
                    return True

        # If the node being modified is just "environ" assume that it's os.environ
        if isinstance(node, astroid.Name):
            if node.name == "environ":
                return True

        return False

    @only_required_for_messages("environment-modify")
    def visit_assign(self, node):
        if not isinstance(node, astroid.Assign):
            return

        # Look for os.environ["WHATEVER"] = something
        for target in node.targets:
            if not isinstance(target, astroid.Subscript):
                continue

            if self._is_environ(target.value):
                self.add_message("environment-modify", node=node)

    @only_required_for_messages("environment-modify")
    def visit_call(self, node):
        # Check both for uses of os.putenv and os.setenv and modifying calls
        # to the os.environ object, such as os.environ.update

        if not isinstance(node, astroid.Call):
            return

        function_node = safe_infer(node.func)
        if not isinstance(function_node, (astroid.FunctionDef, astroid.BoundMethod)):
            return

        # If the function is from the os or posix modules, look for calls that
        # modify the environment
        if function_node.root().name in ("os", os.name) and \
                function_node.name in ("putenv", "unsetenv"):
            self.add_message("environment-modify", node=node)

        # Look for methods bound to the environ dict
        if isinstance(function_node, astroid.BoundMethod) and \
                isinstance(function_node.bound, astroid.Dict) and \
                function_node.bound.root().name in ("os", os.name) and \
                function_node.bound.name == "environ" and \
                function_node.name in ("clear", "pop", "popitem", "setdefault", "update"):
            self.add_message("environment-modify", node=node)

    @only_required_for_messages("environment-modify")
    def visit_delete(self, node):
        if not isinstance(node, astroid.Delete):
            return

        # Look for del os.environ["WHATEVER"]
        for target in node.targets:
            if not isinstance(target, astroid.Subscript):
                continue

            if self._is_environ(target.value):
                self.add_message("environment-modify", node=node)


def register(linter):
    """required method to auto register this checker """
    linter.register_checker(EnvironChecker(linter))



#!/usr/bin/python3

import sys

from pocketlint import PocketLintConfig, PocketLinter


class PocketLintPocketLintConfig(PocketLintConfig):
    @property
    def pylintPlugins(self):
        retval = super(PocketLintPocketLintConfig, self).pylintPlugins
        # We remove things from the environment, but in a safe way since
        # threads are not involved (yet).
        retval.remove("pocketlint.checkers.environ")
        return retval


if __name__ == "__main__":
    conf = PocketLintPocketLintConfig()
    linter = PocketLinter(conf)
    rc = linter.run()
    sys.exit(rc)



# Pango markup pylint module
#
# Copyright (C) 2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Shea <dhsea@redhat.com>
#

import astroid

from pylint.checkers import BaseChecker
from pylint.checkers.utils import only_required_for_messages

import xml.etree.ElementTree as ET

from pocketlint.pangocheck import markup_nodes, is_markup

markupMethods = ["set_markup"]
escapeMethods = ["escape_markup"]

i18n_funcs = ["_", "N_", "P_", "C_", "CN_", "CP_"]
i18n_ctxt_funcs = ["C_", "CN_", "CP_"]


class MarkupChecker(BaseChecker):
    name = "pango-markup"
    msgs = {"W9920" : ("Found invalid pango markup",
                       "invalid-markup",
                       "Pango markup could not be parsed"),
            "W9921" : ("Found pango markup with invalid element %s",
                       "invalid-markup-element",
                       "Pango markup contains invalid elements"),
            "W9922" : ("Found % in markup with unescaped parameters",
                       "unescaped-markup",
                       "Parameters passed to % in markup not escaped"),
           }

    # Check a parsed markup string for invalid tags
    def _validate_pango_markup(self, node, root):
        if root.tag not in markup_nodes:
            self.add_message("W9921", node=node, args=(root.tag,))
        else:
            for child in root:
                self._validate_pango_markup(node, child)

    # Attempt to parse a markup string as XML
    def _validate_pango_markup_string(self, node, string):
        try:
            # QUIS CUSTODIET IPSOS CUSTODES
            # pylint: disable=unescaped-markup
            tree = ET.fromstring("<markup>%s</markup>" % string)
        except ET.ParseError:
            self.add_message("W9920", node=node)
        else:
            # Check that all of the elements are valid for pango
            self._validate_pango_markup(node, tree)

    def __init__(self, linter=None):
        BaseChecker.__init__(self, linter)

    @only_required_for_messages("invalid-markup", "invalid-markup-element", "unescaped-markup", "unnecessary-markup")
    def visit_const(self, node):
        if not isinstance(node.value, (str, bytes)):
            return

        if not is_markup(node.value):
            return

        self._validate_pango_markup_string(node, node.value)

        # Check if this the left side of a % operation
        curr = node
        formatOp = None
        while curr.parent:
            if isinstance(curr.parent, astroid.BinOp) and curr.parent.op == "%" and \
                    curr.parent.left == curr:
                formatOp = curr.parent
                break
            curr = curr.parent

        # Check whether the right side of the % operation is escaped
        if formatOp:
            if isinstance(formatOp.right, astroid.Call):
                if getattr(formatOp.right.func, "name", "") not in escapeMethods:
                    self.add_message("W9922", node=formatOp.right)
            # If a tuple, each item in the tuple must be escaped
            elif isinstance(formatOp.right, astroid.Tuple):
                for elt in formatOp.right.elts:
                    if not isinstance(elt, astroid.Call) or\
                            getattr(elt.func, "name", "") not in escapeMethods:
                        self.add_message("W9922", node=elt)
            # If a dictionary, each value must be escaped
            elif isinstance(formatOp.right, astroid.Dict):
                for item in formatOp.right.items:
                    if not isinstance(item[1], astroid.Call) or\
                            getattr(item[1].func, "name", "") not in escapeMethods:
                        self.add_message("W9922", node=item[1])
            else:
                self.add_message("W9922", node=formatOp)


def register(linter):
    """required method to auto register this checker """
    linter.register_checker(MarkupChecker(linter))



# Base classes for pocketlint
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Chris Lumens <clumens@redhat.com>
#

from __future__ import print_function

import atexit
import concurrent.futures
import os
import re
import shutil
import subprocess
import sys
import tempfile

from packaging.version import Version


class PocketLintConfig(object):
    """Configuration object that a project should use to tell pylint how
       to operate.  Instance attributes:

       falsePositives    -- A list of FalsePositive objects for filtering
                            incorrect pylint error messages.
                            Default: []
       loadAllExtensions -- Load all the extensions to the pylint. This will resolve
                            I1101(c-extension-no-member) messages but it could be
                            potentially dangerous. See man pylint --unsafe-load-any-extension
                            Default: False
    """
    def __init__(self):
        self.falsePositives = []
        self.loadAllExtensions = False

    @property
    def disabledOptions(self):
        """A list of warning and error codes that pylint should ignore
           when checking source code.  The base PocketLintConfig object
           comes with a pretty useful list, but subclasses can feel free
           to add or remove as desired.  This list should strive to be empty,
           though.
        """
        return  ["W0123",           # Use of eval
                 "W0212",           # Access to a protected member %s of a client class
                 "W0511",           # Used when a warning note as FIXME or XXX is detected
                 "W0603",           # Using the global statement
                 "W0613",           # Unused argument %r
                 "W0614",           # Unused import %s from wildcard import
                 "E0012",           # Bad inline option given to pylint
                 "I0011",           # Locally disabling %s
                 "I0013",           # Ignoring entire file
               ]

    @property
    def extraArgs(self):
        """Extra command line arguments that should be passed to pylint.  These
           arguments will be added after the default ones so they can override
           the base config, but may in turn also be overridden by arguments
           passed to the testing framework on the command line.
        """
        return []

    @property
    def initHook(self):
        """Python code to be run by pylint as part of an init hook.  Most
           projects will not need this.
        """
        return ""

    @property
    def pylintPlugins(self):
        """A list of plugins provided by PocketLint to be added to the set of
           pylink checkers.  Not all of these will be relevant to all projects,
           but they should still be able to run without error.  If necessary,
           projects can modify this list as needed.
        """
        return [ "pocketlint.checkers.environ",
                 "pocketlint.checkers.intl",
                 "pocketlint.checkers.markup",
                 "pocketlint.checkers.pointless-override",
                 "pocketlint.checkers.preconf",
               ]

    @property
    def ignoreNames(self):
        """A set of names to skip when automatically determining the list of
           files to lint. The items in this set could be a particular filename
           to skip or a directory that the linter should not traverse into.
           The items should be just basenames, not paths.
        """
        return set()


class FalsePositive(object):
    """An object used in filtering out incorrect results from pylint.  Pass in
       a regular expression matching a pylint error message that should be
       ignored.  This object can also be used to keep track of how often it is
       used, for auditing that false positives are still useful.
    """
    def __init__(self, regex):
        self.regex = regex
        self.used = 0


class PocketLinter(object):
    """Main class that does the hard work of running pylint on a project.
       Pass an instance of PocketLintConfig to a new instance of this class
       and then call its run method.  This is all that should be necessary for
       most projects:

       from pocketlint import PocketLintConfig, PocketLinter

       class FooLintConfig(PocketLintConfig):
          ....

       if __name__ == "__main__":
           conf = FooLintConfig()
           linter = PocketLinter(conf)
           rc = linter.run()
           os._exit(rc)
    """
    def __init__(self, config):
        self._config = config
        self._pylint_log = False

        # If top_srcdir is set, assume this is being run from automake and we don't
        # need to keep a separate log.
        if "top_srcdir" not in os.environ:
            self._pylint_log = True

    def _del_xdg_runtime_dir(self):
        shutil.rmtree(os.environ["XDG_RUNTIME_DIR"])

    @property
    def _files(self):
        retval = []

        srcdir = os.environ.get("top_srcdir", os.getcwd())

        for (root, dirnames, files) in os.walk(srcdir):
            # Filter out the names to ignore
            for i in self._config.ignoreNames:
                if i in dirnames:
                    dirnames.remove(i)
                if i in files:
                    files.remove(i)

            for f in files:
                try:
                    with open(root + "/" + f, encoding='utf-8') as fo:
                        lines = fo.readlines()
                except UnicodeDecodeError:
                    # If we couldn't open this file, just skip it.  It wasn't
                    # going to be valid python anyway.
                    continue

                if "# pylint: skip-file\n" in lines:
                    continue

                # Test any file that either ends in .py or contains #!/usr/bin/python
                # in the first line.
                if f.endswith(".py") or (lines and str(lines[0]).startswith("#!/usr/bin/python")):
                    retval.append(root + "/" + f)

        return retval

    @property
    def _pylint_args(self):
        args = [ "--msg-template='{msg_id}({symbol}):{line:3d},{column}: {obj}: {msg}'",
                 "-r", "n",
                 "--disable", "C,R",
                 "--rcfile", "/dev/null",
                 "--dummy-variables-rgx", "_",
                 "--ignored-classes", "DefaultInstall,Popen,QueueFactory,TransactionSet",
                 "--defining-attr-methods", "__init__,grabObjects,initialize,reset,start,setUp",
                 "--load-plugins", ",".join(self._config.pylintPlugins),
                 "--deprecated-modules", "string,regsub,TERMIOS,Bastion,rexec",
                 "--disable", ",".join(self._config.disabledOptions),
               ]

        if self._config.initHook:
            args += ["--init-hook", self._config.initHook]

        if self._config.extraArgs:
            args += self._config.extraArgs

        if self._config.loadAllExtensions:
            args += ["--unsafe-load-any-extension=yes"]

        # since 1.7 pylint by default prints "score", we need to disable it
        if self._pylint_version >= Version("1.7.0"):
            args += ["-s", "n"]

        return args

    def _command_exists(self, exc):
        proc = subprocess.Popen(["which", exc],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        _out, _err = proc.communicate()
        return proc.returncode == 0

    @property
    def _pylint_executable(self):
        return [sys.executable, "-m", "pylint"]

    @property
    def _pylint_version(self):
        exc = self._pylint_executable
        exc.append("--version")
        proc = subprocess.Popen(exc, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (stdout, _stderr) = proc.communicate()
        pattern = re.compile(r".+ (?P<version>[0-9.]+)")
        match = pattern.search(stdout.decode())
        if match:
            return Version(match.group("version"))
        else:
            return Version("0")

    def _parseArgs(self):
        # Really stupid argument processing - strip off the first argument (that's
        # the program name) and then anything starting with a dash is an argument to
        # pylint, and anything else is a file to check.  We don't want this program
        # to have much more in the way of argument processing because the user should
        # do configuration via a LintConfig object.
        args = []
        files = []

        if len(sys.argv) > 1:
            for arg in sys.argv[1:]:
                if arg.startswith("-"):
                    args.append(arg)
                else:
                    files.append(arg)

        return (args, files)

    def _setupEnvironment(self):
        # We need top_builddir to be set so we know where to put the pylint analysis
        # stuff.  Usually this will be set up if we are run via "make test" but if
        # not, hope that we are at least being run out of the right directory.
        builddir = os.environ.get("top_builddir", os.getcwd())

        # XDG_RUNTIME_DIR is "required" to be set, so make one up in case something
        # actually tries to do something with it.
        if "XDG_RUNTIME_DIR" not in os.environ:
            d = tempfile.mkdtemp()
            os.environ["XDG_RUNTIME_DIR"] = d
            atexit.register(self._del_xdg_runtime_dir)

        # Unset TERM so that things that use readline don't output terminal garbage.
        if "TERM" in os.environ:
            os.environ.pop("TERM")

        # Don't try to connect to the accessibility socket.
        os.environ["NO_AT_BRIDGE"] = "1"

        # Force the GDK backend to X11.  Otherwise if no display can be found, Gdk
        # tries every backend type, which includes "broadway", which prints an error
        # and keeps changing the content of said error.
        os.environ["GDK_BACKEND"] = "x11"

        # Save analysis data in the pylint directory.
        os.environ["PYLINTHOME"] = builddir + "/tests/pylint/.pylint.d"
        if not os.path.exists(os.environ["PYLINTHOME"]):
            os.mkdir(os.environ["PYLINTHOME"])

    def _filterFalsePositives(self, filename, lines):
        if not self._config.falsePositives:
            return lines

        retval = []

        for line in lines:
            # This is not an error message.  Ignore it.
            if line.startswith("*****") or line.startswith("Using config file") or not line.strip():
                continue
            else:
                validError = True

                for regex in self._config.falsePositives:
                    if re.search(regex.regex, line):
                        # The false positive was hit, so record that and ignore
                        # the message from pylint.
                        regex.used += 1
                        validError = False
                        break

                # If any false positive matched the error message, it's a valid
                # error from pylint.  Add it to what we're going to return.
                if validError:
                    retval.append(line)

        # If any errors were found, add the header line with the name of the file
        # (instead of the name of the module) and return the list.
        if retval:
            retval.insert(0, "************* Module " + filename)

        return retval

    def _run_one(self, filename, args):
        proc = subprocess.Popen(self._pylint_executable + self._pylint_args + args + [filename],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (stdout, stderr) = proc.communicate()
        output = stdout + stderr

        lines = self._filterFalsePositives(filename, output.decode("utf-8").split("\n"))
        if lines:
            return ("\n".join(lines), proc.returncode)

        return ("", 0)

    def _print(self, s, fo=None):
        print(s)
        sys.stdout.flush()
        if fo:
            print(s, file=fo)
            fo.flush()

    def run(self):
        retval = 0

        self._setupEnvironment()

        (args, files) = self._parseArgs()

        if not files:
            files = self._files

        if self._pylint_log:
            fo = open("pylint-log", "w", encoding='utf-8')
        else:
            fo = None

        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            jobs = []
            for f in files:
                jobs.append(executor.submit(self._run_one, f, args))

            for job in concurrent.futures.as_completed(jobs):
                result = job.result()

                output = result[0].strip()
                if output:
                    self._print(output, fo)

                if result[1] > retval:
                    retval = result[1]

        unusedFPs = []

        for fp in self._config.falsePositives:
            if fp.used == 0:
                unusedFPs.append(fp.regex)

        if unusedFPs:
            self._print("************* Unused False Positives Found:", fo)

            for fp in unusedFPs:
                self._print(fp, fo)

        return retval



#!/usr/bin/python3

from setuptools import setup

setup(name='pocketlint', version='0.25',
      description='Support for running pylint against projects',
      author='Chris Lumens', author_email='clumens@redhat.com',
      url='https://github.com/rhinstaller/pocketlint',
      license='COPYING',
      install_requires=['pylint', 'polib', 'packaging'],
      long_description=open('README', encoding='utf-8').read(),
      packages=['pocketlint', 'pocketlint.checkers'],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Testing",
      ],
      )
